/* R-Dev-Ino dual Motor Controller (slave only)
   - uses I2C or Serial communication interfaces
   - uses a dual H-bridge to drive 2 DC motors
   - uses quadrature encoders to get feedback from the motors, 
   single interrupt mode, dual count (141x2=282CPR)
   - uses a trapezoidal profile to accelerate and decelerate
   - will use a PID control algorithm to keep a straight line
   - master has to keep track of traveled distance
*/ 

/* Distance calculations, specific to my robot:
  wheel diameter = 81.6mm -> wheel circumference = 81.6 x 3.1416 = 256.35456 mm
  distance between wheels = 160 mm -> turn circumference = 160 x 3.1416 = 502.656 mm
  mm / turn degree = 502.656 / 360 = 1.396
  motor RPM = 80 -> RPS = 80/60 = 1.3333
  encoder clics / wheel rotation = 282
  mm / click = 256.3545 / 282 = 0.9090
  clicks / mm = 282 / 256.35456 = 1.1
  clicks / turn deg = 1.396 / 1.1 = 1.269 

 For standard servo wheels:
  wheel dia = 66.8mm x 3.1416 => wheel circumference = 210
  wheel dist = 170 => turn circumference = 534.072
  mm/deg = 471.24/360=1.483
  clicks/mm = 282/210=1.34
  clicks/deg = 1.483*1.34=1.98722
*/

/* Driving and Turning with trapezoidal control profile
 For a given distance, we ramp up to set speed, drive or turn at a constant speed,
 then we ramp down to a stop; if the distance is too short, we ramp up to the middle of
 the distance then we ramp down to a stop.
*/

// R-Dev-Ino board pinout:
//
// Ext  RDI Funct Arduino  ATmega168      Arduino Funct RDI External
//                       +-----\/----+
//          Reset       1| PC6   PC5 |28  D19 A5  SCL
//          Rx    D0    2| PD0   PC4 |27  D18 A4  SDA
//          Tx    D1    3| PD1   PC3 |26  D17 A3        
//      E1A Int0  D2    4| PD2   PC2 |25  D16 A2        
//      E2A Int1  D3    5| PD3   PC1 |24  D15 A1        
//      E1B       D4    6| PD4   PC0 |23  D14 A0        
//                      7| VCC   GND |22  
//                      8| GND  AREF |21  
//          Xtal        9| PB6  AVCC |20  
//          Xtal       10| PB7   PB5 |19  D13      SCK  LED
//      E2B OC0B  D5   11| PD5   PB4 |18  D12      MISO M1B
//          OC0A  D6   12| PD6   PB3 |17  D11 OC2A MOSI       
//                D7   13| PD7   PB2 |16  D10 OC1B      M1A
//      M2B       D8   14| PB0   PB1 |15  D 9 OC1A      M2A
//                       +-----------+
//

#include <Wire.h>

//Inputs/outputs
#define Encoder_1_ChA  2 // digital pin 2    // Left Encoder
#define Encoder_1_ChB  4 // digital pin 4
#define Encoder_2_ChA  3 // digital pin 3    // Right Encoder
#define Encoder_2_ChB  5 // digital pin 5

#define Motor_1_PWM   9 // digital pin 9    // Right Motor
#define Motor_1_Dir  12 // digital pin 12
#define Motor_2_PWM  10 // digital pin 10   // Left Motor
#define Motor_2_Dir   8 // digital pin 8

//Variables
byte CpCm = 134;   // clicks per centimeter (13.4x10)
byte CmMf = 10;     // cm multiplying factor
byte CpDg = 198;    // clicks per degree (turn) (1.75x100)
byte DgMf = 100;    // degree multiplying factor

byte Acc = 10;
byte dir = 1;
byte x=0;
byte y=0;
byte a=0;
boolean stopNow=0;
char Error=0;

volatile int Encoder_Count[2]={0,0};
#define E1 0
#define E2 1

byte I2C[5]={0,0,0,0,0}; 
#define Address   0
#define Command   1
#define Value1    2
#define Value2    3

// I2C addresses
#define Brain           20>>1
#define Compass         42>>1
#define Speech          50>>1
#define MotorController 60>>1
#define ServoController 70>>1
#define LedBar          80>>1
#define Mapper          90>>1
#define Thermo          208>>1

// I2C commands
#define Forward     1 //MotorController
#define Reverse     2 //MotorController
#define Left        3 //MotorController/ServoController
#define Right       4 //MotorController/ServoController
#define StopAll     5 //MotorController
#define GoFwd       6 //MotorController
#define GoBwd       7 //MotorController
#define Err         8 //MotorController

#define Up          1 //ServoController
#define Down        2 //ServoController

#define Swaist     11 //ServoController
#define SleftSh    12 //ServoController
#define SrightSh   13 //ServoController
#define SleftArm   14 //ServoController
#define SrightArm  15 //ServoController
#define Sscan      16 //ServoController
#define SleftHand  17 //ServoController
#define SrightHand 18 //ServoController
#define SpanHead   19 //ServoController
#define StiltHead  20 //ServoController
#define StoreMove  21 //ServoController
#define PickUp     22 //ServoController
#define Release    23 //ServoController
#define Wave       24 //ServoController

#define LCD1       31 //Speech
#define LCD2       32 //Speech
#define Heading    33 //Speech
#define Node       34 //Speech
#define Room       35 //Speech
#define Ping       36 //Speech
#define Sharp      37 //Speech
#define Angle      38 //Speech
#define Compass    39 //Speech
#define Temp       40 //Speech
#define Cluster    41 //Speech
#define Fire       42 //Speech

#define Scan       61 //Mapper
#define Fire       62 //Mapper
#define Fetch      63 //Mapper
#define Dance      64 //Mapper

#define Done       99 //Mapper

#define ResetCounters() Encoder_Count[E1] = 0; Encoder_Count[E2] = 0
#define kp 20
//-----------------------------------------------------------------------------
void setup() { 
   
  pinMode(Encoder_1_ChA, INPUT); 
   //digitalWrite(Encoder_1_ChA, HIGH);       // turn on pullup resistor
  pinMode(Encoder_1_ChB, INPUT); 
   //digitalWrite(Encoder_1_ChB, HIGH);       // turn on pullup resistor
  pinMode(Encoder_2_ChA, INPUT); 
   //digitalWrite(Encoder_2_ChA, HIGH);       // turn on pullup resistor
  pinMode(Encoder_2_ChB, INPUT); 
   //digitalWrite(Encoder_2_ChB, HIGH);       // turn on pullup resistor

  // set motor pins as output and LOW so the motors are breaked
  pinMode(Motor_1_PWM, OUTPUT);
  pinMode(Motor_1_Dir, OUTPUT);
  pinMode(Motor_2_PWM, OUTPUT);
  pinMode(Motor_2_Dir, OUTPUT);
  Braked();
  attachInterrupt(0, doEncoder_1, CHANGE);  // encoder pin on interrupt 0
  attachInterrupt(1, doEncoder_2, CHANGE);  // encoder pin on interrupt 1

  //Serial.begin (9600);
  //Serial.println("start");

  Wire.begin(MotorController);     // join i2c bus with address #60
  Wire.onReceive(I2CreceiveEvent); // register event

} 

//-----------------------------------------------------------------------------
void loop(){
  Process_commands();
  //DriveTest();
}  
//-----------------------------------------------------------------------------

void Process_commands() {
    switch (I2C[Command]) {
      case 0:
        break;
      case Forward:
        dir = 1;
        //Serial.println("Moving Forward");
        Drive(I2C[Value1], I2C[Value2]); // distance, speed
        I2C_transmit(Mapper, Done, 0, 0);
        I2C[Command]=0;
        break;
      case Reverse:
        dir = 0;
        //Serial.println("Moving Backward");
        Drive(I2C[Value1], I2C[Value2]); // distance, speed
        I2C_transmit(Mapper, Done, 0, 0);
        I2C[Command]=0;
        break;
      case Left:
        dir = 2;
        //Serial.println("Turning Left");
        Turn(I2C[Value1], I2C[Value2]); // angle, speed 
        I2C_transmit(Mapper, Done, 0, 0);
        I2C[Command]=0;
        break;
      case Right:
        dir = 3;
        //Serial.println("Turning Right");
        Turn(I2C[Value1], I2C[Value2]); // angle, speed 
        I2C_transmit(Mapper, Done, 0, 0);
        I2C[Command]=0;
        break;
      case StopAll:
        //Serial.println("Braked!");
        Braked();
        I2C[Command]=0;
        break;
      case GoFwd: // Go Forward
        //Serial.println("Driving Forward");
          digitalWrite(Motor_1_Dir, LOW); // forward
          analogWrite(Motor_1_PWM, I2C[Value2]); // Right speed
          digitalWrite(Motor_2_Dir, LOW); // forward
          analogWrite(Motor_2_PWM, I2C[Value1]); // Left speed
        I2C[Command]=0;
        break;
      case GoBwd: // Go Backward
        //Serial.println("Driving Backward");
          digitalWrite(Motor_1_Dir, HIGH); // forward
          analogWrite(Motor_1_PWM, 255-I2C[Value2]); // Right speed
          digitalWrite(Motor_2_Dir, HIGH); // forward
          analogWrite(Motor_2_PWM, 255-I2C[Value1]); // Left speed
        I2C[Command]=0;
        break;
    }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void Drive(int Distance, int Speed){
  if (Distance > 0){
    int sr=0, sl=0, error=0;
    int Clicks = 0;
    int D=0;
    int ConstSpeedDistance = 0;
    boolean Stopped=0;
    Distance = Distance * CpCm / CmMf; // convert distance from cm to clicks
    D=Distance;
    //Serial.print ("Distance: ");            // debug - remember to comment out
    //Serial.println (D, DEC);        // debug - remember to comment out
    ResetCounters();
    for (int x=0; x<=Speed; x=x+Acc) {
      if (!stopNow){
      Clicks = Encoder_Count[E1]; // read the encoders
      //Serial.print ("Clicks: ");            // debug - remember to comment out
      //Serial.println (Clicks, DEC);        // debug - remember to comment out
      if (Clicks >= Distance/2 && !stopNow){  // we don't have time to ramp up to max speed
        Stop(D,x);                  // so we start the deceleration
        x = Speed + Acc;          // set x so we get out of the for loop
        //Clicks = Encoder_Count[E1]; // read the encoders
        //delay(1);
        //Serial.print ("Clicks4: ");            // debug - remember to comment out
        //Serial.println (Clicks, DEC);        // debug - remember to comment out
        Stopped=1;
      }//end if
      if (!Stopped&&!stopNow){
        if (x > Speed-Acc) {        // if we're close to the set speed
          ConstSpeedDistance = Distance - Clicks; // set the distance for constant speed
          //Serial.print ("CSD: ");            // debug - remember to comment out
          //Serial.println (ConstSpeedDistance, DEC);        // debug - remember to comment out
        }//end if
      }//end if
      // adjust left motor PWM so it keeps up with the right motor
      error=Encoder_Count[E1]-Encoder_Count[E2];
      sr=x;
      sl=x+kp*error/10+Error;
      // update the motors PWM
      if (dir == 1){ // forward
        digitalWrite(Motor_1_Dir, LOW); // forward
        digitalWrite(Motor_2_Dir, LOW); // forward
        analogWrite(Motor_1_PWM, sr); // direct PWM, low values mean low speed
        analogWrite(Motor_2_PWM, sl); //
      }//end if
      if (dir == 0){ // reverse
        digitalWrite(Motor_1_Dir, HIGH); // reverse
        digitalWrite(Motor_2_Dir, HIGH); // reverse
        analogWrite(Motor_1_PWM, 255-sr); // inverse PWM, low values mean high speed
        analogWrite(Motor_2_PWM, 255-sl); //
      }//end if
      Error=0;    
      delay(15);
    }//end if !stopNow
    }//end for
    if (!Stopped&&!stopNow){
      while (ConstSpeedDistance>Clicks&&!stopNow) { // keep driving at set speed
        Clicks = Encoder_Count[E1]; // read the encoders
        // adjust left motor PWM so it keeps up with the right motor
        error=Encoder_Count[E1]-Encoder_Count[E2];
        sr=Speed;
        sl=Speed+kp*error/10+Error;
        if (sl>255){sl=255;} //don't go over the max speed
        // update the motors PWM
        if (dir == 1){ // forward
          analogWrite(Motor_1_PWM, sr); // direct PWM, low values mean low speed
          analogWrite(Motor_2_PWM, sl); //
        }//end if
        if (dir == 0){ // reverse
          analogWrite(Motor_1_PWM, 255-sr); // inverse PWM, low values mean high speed
          analogWrite(Motor_2_PWM, 255-sl); //
        }//end if    
      Error=0;    
      }//end while 
      Stop(D, Speed);                // so we start the deceleration
    }//end if
    Clicks = Encoder_Count[E1]; // read the encoders
    //Serial.print ("Clicks3: ");            // debug - remember to comment out
    //Serial.println (Clicks, DEC);        // debug - remember to comment out
  }else{
    if (dir == 1){ // forward
      digitalWrite(Motor_1_Dir, LOW); // forward
      digitalWrite(Motor_2_Dir, LOW); // forward
      analogWrite(Motor_1_PWM, Speed); // direct PWM, low values mean low speed
      analogWrite(Motor_2_PWM, Speed); //
    }//end if
    if (dir == 0){ // reverse
      digitalWrite(Motor_1_Dir, HIGH); // reverse
      digitalWrite(Motor_2_Dir, HIGH); // reverse
      analogWrite(Motor_1_PWM, 255-Speed); // inverse PWM, low values mean high speed
      analogWrite(Motor_2_PWM, 255-Speed); //
    }//end if    
  }
}//end void

void Turn(int Angle, int Speed) {
  if (Angle>0){
    int Clicks = 0;
    int A=0;
    int ConstSpeedTurn = 0;
    boolean Stopped=0;
    Angle = Angle * CpDg / DgMf; // convert Angle from degrees to clicks
    //Serial.print ("Angle: ");            // debug - remember to comment out
    //Serial.println (Angle, DEC);        // debug - remember to comment out
    ResetCounters();
    for (int x=0; x<=Speed; x=x+Acc) {
      Clicks = (Encoder_Count[E1]+Encoder_Count[E1])/2; // read the encoders
      //Serial.print ("Clicks: ");            // debug - remember to comment out
      //Serial.println (Clicks, DEC);        // debug - remember to comment out
      if (Clicks >= Angle/2){  // we don't have time to ramp up to max speed
        Stop(A, x);               // so we start the deceleration
        x = Speed + Acc;       // set L so we get out of the for loop
        //Clicks = Encoder_Count[E1]; // read the encoders
        //Serial.print ("Clicks4: ");            // debug - remember to comment out
        //Serial.println (Clicks, DEC);        // debug - remember to comment out
        Stopped=1;
      }
      if (!Stopped){
        if (x > Speed-Acc) {      // if we're close to the set speed
          ConstSpeedTurn = Angle - Clicks; // set the angle for constant speed
          //Serial.print ("CST: ");            // debug - remember to comment out
          //Serial.println (ConstSpeedTurn, DEC);        // debug - remember to comment out
        }
      }
      // update the motors PWM
      if (dir == 2){ // right
        digitalWrite(Motor_1_Dir, HIGH); // reverse
        digitalWrite(Motor_2_Dir, LOW); // forward
        analogWrite(Motor_1_PWM, 255-x); // inverse PWM, low values mean high speed
        analogWrite(Motor_2_PWM, x); // direct PWM, low values mean low speed
      }
      if (dir == 3){ // left
        digitalWrite(Motor_1_Dir, LOW); // forward
        digitalWrite(Motor_2_Dir, HIGH); // reverse
        analogWrite(Motor_1_PWM, x); // direct PWM, low values mean low speed
        analogWrite(Motor_2_PWM, 255-x); // inverse PWM, low values mean high speed
      }
      delay(15);
    }//end for
    if (!Stopped){
      while (ConstSpeedTurn > Clicks) {
        Clicks = (Encoder_Count[E1]+Encoder_Count[E1])/2; // read the encoders
      }
      Stop(A, Speed);
    }  
    //Clicks = Encoder_Count[E1]; // read the encoders
    //Serial.print ("Clicks3: ");            // debug - remember to comment out
    //Serial.println (Clicks, DEC);        // debug - remember to comment out
  }else{
    if (dir == 2){ // right
      digitalWrite(Motor_1_Dir, HIGH); // reverse
      digitalWrite(Motor_2_Dir, LOW); // forward
      analogWrite(Motor_1_PWM, 255-Speed); // inverse PWM, low values mean high speed
      analogWrite(Motor_2_PWM, Speed); // direct PWM, low values mean low speed
    }
    if (dir == 3){ // left
      digitalWrite(Motor_1_Dir, LOW); // forward
      digitalWrite(Motor_2_Dir, HIGH); // reverse
      analogWrite(Motor_1_PWM, Speed); // direct PWM, low values mean low speed
      analogWrite(Motor_2_PWM, 255-Speed); // inverse PWM, low values mean high speed
    }
  }
}

void Stop(int Dist, int Speed){ // decelerate from set speed to stop
  int Clicks = 0;
  int s=0;
  Speed=Speed-40; 
  Clicks = (Encoder_Count[E1] + Encoder_Count[E2])/2; // read the encoders
    //Serial.print ("Dist: ");            // debug - remember to comment out
    //Serial.println (Dist, DEC);        // debug - remember to comment out
  while (Clicks < Dist) {
     if (dir==0){// dir = reverse
       digitalWrite(Motor_1_Dir, HIGH); // reverse
       digitalWrite(Motor_2_Dir, HIGH); // reverse
       analogWrite(Motor_1_PWM, 255-Speed+s); // inverse PWM, low values mean high speed
       analogWrite(Motor_2_PWM, 255-Speed+s); // inverse PWM, low values mean high speed
     }
     if (dir==1){ // dir = forward
       digitalWrite(Motor_1_Dir, LOW); // forward
       digitalWrite(Motor_2_Dir, LOW); // forward
       analogWrite(Motor_1_PWM, Speed-s);
       analogWrite(Motor_2_PWM, Speed-s);
     }
     if (dir==2){ // dir = left
       digitalWrite(Motor_1_Dir, HIGH); // reverse
       digitalWrite(Motor_2_Dir, LOW); // forward
       analogWrite(Motor_1_PWM, 255-Speed+s); 
       analogWrite(Motor_2_PWM, Speed-s); 
     }
     if (dir==3){ // dir = right
       digitalWrite(Motor_1_Dir, LOW); // forward
       digitalWrite(Motor_2_Dir, HIGH); // reverse
       analogWrite(Motor_1_PWM, Speed-s); 
       analogWrite(Motor_2_PWM, 255-Speed+s); // 
     }  
     if (s<=Speed){ s=s+Acc;}
     delay(3);
     Clicks = (Encoder_Count[E1] + Encoder_Count[E2])/2; // read the encoders
     //Serial.print ("Clicks_end: ");            // debug - remember to comment out
     //Serial.println (Clicks, DEC);        // debug - remember to comment out
   }
   if (Clicks >= Dist){ Braked();}
}

void Braked()
{
  digitalWrite(Motor_1_PWM, LOW);
  digitalWrite(Motor_1_Dir, LOW);
  digitalWrite(Motor_2_PWM, LOW);
  digitalWrite(Motor_2_Dir, LOW);
}  

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void doEncoder_1(){
  Encoder_Count[E1]++;         // CCW
  //Serial.print ("E1 count: ");            // debug - remember to comment out
  //Serial.println (Encoder_Count[E1], DEC);        // debug - remember to comment out
}

void doEncoder_2(){
  Encoder_Count[E2]++;
  //Serial.print ("E2 count: ");            // debug - remember to comment out
  //Serial.println (Encoder_Count[E2], DEC);        // debug - remember to comment out
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void I2CreceiveEvent(int i)
{
    //Serial.println("I2C communication received:");
    while(0 < Wire.available()) // loop through all
    {
      I2C[i] = Wire.receive(); // receive bytes
      //Serial.print(i, DEC);
      //Serial.print("  ");
      //Serial.println(I2C[i], DEC);
      i = i-1;
    }
   // if (I2C[Command]==StopAll) stopNow=1;
   // if (I2C[Command]==Err) Error=char(I2C[Value1]);
}

void I2C_transmit(byte Addr, byte Cmnd, byte Val1, byte Val2){
  Wire.beginTransmission(Addr); // transmit to slave device
  Wire.send(Val2);         // sends one byte  
  Wire.send(Val1);         // sends one byte  
  Wire.send(Cmnd);        // sends one byte
  Wire.endTransmission();    // stop transmitting
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void DriveTest(){
  dir = 1;
  Drive(50,200);
  delay(3000);
  dir = 3;
  Turn(90,150);
  delay(3000);
}
